/* A demo of signal-based periodic execution of tasks.
 *
 * Compilation:
 * cc -lrt rt-signal-demo.c -o rt-signal-demo && ./rt-signal-demo
 *
 * However, note that [1] says that you should not use the signal mechanism for
 * realtime code. Still, we give this demo for illustration purpose.
 *
 * We do not perform all the memory tweaking (e.g., page locking). See
 * rt-nanosleep-demo instead.
 *
 * We do not actually set a real-time scheduling policy here. See
 * rt-nanosleep-demo instead.
 *
 * Further information:
 * [1] Marta Rybczyńska, A realtime developer's checklist,
 *     https://lwn.net/Articles/837019/
 * [2] man time_create, in particular see the code example
 */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <signal.h>
#include <time.h>


static void handler(int sig, siginfo_t *si, void *uc ) {
    static unsigned cnt = 0;
    ++cnt;

    if (cnt % 10 == 0)
        putchar('|');
    else
        putchar('-');

    fflush(stdout);
}


int main(int argc, char *argv[]) {
	struct sigevent sev;
	struct itimerspec its;
	long long freq_nanosecs;
	sigset_t mask;
	struct sigaction sa;

	printf("Establishing handler for signal %d\n", SIGRTMIN);
	sa.sa_flags = SA_SIGINFO;
	sa.sa_sigaction = handler;
	sigemptyset(&sa.sa_mask);
	if (sigaction(SIGRTMIN, &sa, NULL) == -1) {
        perror("sigaction failed");
        return EXIT_FAILURE;
    }

    timer_t timerid;

	/* Create the timer */
	sev.sigev_notify = SIGEV_SIGNAL;
	sev.sigev_signo = SIGRTMIN;
	sev.sigev_value.sival_ptr = &timerid;

    /* In contast to CLOCK_REALTIME, CLOCK_MONOTONIC can not jump around
     * (including backward) through system clock updates (e.g., due to NTP).
     * Use CLOCK_MONOTONIC for realtime code, see [1]. */
	if (timer_create(CLOCK_MONOTONIC, &sev, &timerid) == -1) {
		perror("timer_create failed");
        exit(EXIT_FAILURE);
    }

	printf("timer ID is 0x%lx\n", (long) timerid);

    /* Initial expiration must be non-zero. */
	its.it_value.tv_sec = 0;
	its.it_value.tv_nsec = 1;
    /* Interval of follow-up expirations every 100 ms. */
	its.it_interval.tv_sec = 0;
	its.it_interval.tv_nsec = 100000000;

    if (timer_settime(timerid, 0, &its, NULL) == -1) {
        perror("timer_settime failed");
        exit(EXIT_FAILURE);
    }

    /* The sleeping beauty */
    while (1)
        sleep(10);

	return EXIT_SUCCESS;
}

